"""Module to manage data in a SQLite database."""

import aiosqlite
import asyncio
import contextvars
import fondat.codec
import fondat.sql
import functools
import logging
import sqlite3
import types
import typing
import uuid

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from fondat.codec import Codec, DecodeError, String
from fondat.sql import Expression, Param
from fondat.types import is_optional, is_subclass, literal_values, split_annotated
from fondat.validation import validate_arguments
from types import NoneType
from typing import Any, Literal


_logger = logging.getLogger(__name__)


class SQLiteCodec(Codec[fondat.codec.F, Any]):
    """Base class for SQLite codecs."""


codec_providers = []


@functools.cache
def get_codec(python_type: Any) -> SQLiteCodec:
    """Return a codec compatible with the specified Python type."""

    python_type, _ = split_annotated(python_type)

    for provider in codec_providers:
        if (codec := provider(python_type)) is not None:
            return codec

    raise TypeError(f"failed to provide SQLite codec for {python_type}")


def _codec_provider(wrapped=None):
    if wrapped is None:
        return functools.partial(_codec_provider)
    codec_providers.append(wrapped)
    return wrapped


@_codec_provider
def _blob_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a value to/from a SQLite BLOB. Supports the
    following types: bytes, bytearray.
    """

    if not is_subclass(python_type, (bytes, bytearray)):
        return

    class BlobCodec(SQLiteCodec[python_type]):

        sql_type = "BLOB"

        @validate_arguments
        def encode(self, value: python_type) -> bytes:
            return bytes(value)

        @validate_arguments
        def decode(self, value: bytes) -> python_type:
            return python_type(value)

    return BlobCodec()


@_codec_provider
def _integer_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a value to/from a SQLite INTEGER. Supports the
    following types: int, bool.
    """

    if not is_subclass(python_type, int):  # includes bool
        return

    class IntegerCodec(SQLiteCodec[python_type]):

        sql_type = "INTEGER"

        @validate_arguments
        def encode(self, value: python_type) -> int:
            return int(value)

        @validate_arguments
        def decode(self, value: int) -> python_type:
            return python_type(value)

    return IntegerCodec()


@_codec_provider
def _real_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a value to/from a SQLite REAL. Supports the
    following type: float.
    """

    if not is_subclass(python_type, float):
        return

    class RealCodec(SQLiteCodec[python_type]):

        sql_type = "REAL"

        @validate_arguments
        def encode(self, value: python_type) -> float:
            return float(value)

        @validate_arguments
        def decode(self, value: float) -> python_type:
            return python_type(value)

    return RealCodec()


@_codec_provider
def _union_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a UnionType, Union or Optional value to/from a
    compatible SQLite value. For an optional type, it will use the codec for its type,
    otherwise it will encode/decode as TEXT.
    """

    origin = typing.get_origin(python_type)
    if origin not in {typing.Union, types.UnionType}:
        return

    args = typing.get_args(python_type)
    is_nullable = is_optional(python_type)
    args = [a for a in args if a is not NoneType]
    codec = get_codec(args[0]) if len(args) == 1 else _text_codec_provider(python_type)

    class UnionCodec(SQLiteCodec[python_type]):

        sql_type = codec.sql_type

        @validate_arguments
        def encode(self, value: python_type) -> Any:
            if value is None:
                return None
            return codec.encode(value)

        def decode(self, value: Any) -> python_type:
            if value is None and is_nullable:
                return None
            return codec.decode(value)

    return UnionCodec()


@_codec_provider
def _literal_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a Literal value to/from a compatible SQLite value.
    If all literal values share the same type, then it will use a codec for that type,
    otherwise it will encode/decode as TEXT.
    """

    origin = typing.get_origin(python_type)

    if origin is not Literal:
        return

    literals = literal_values(python_type)
    types = list({type(literal) for literal in literals})
    codec = get_codec(types[0]) if len(types) == 1 else _text_codec_provider(python_type)
    is_nullable = is_optional(python_type) or None in literals

    class LiteralCodec(SQLiteCodec[python_type]):

        sql_type = codec.sql_type

        @validate_arguments
        def encode(self, value: python_type) -> Any:
            if value is None:
                return None
            return codec.encode(value)

        def decode(self, value: Any) -> python_type:
            if value is None and is_nullable:
                return None
            result = codec.decode(value)
            if result not in literals:
                raise DecodeError
            return result

    return LiteralCodec()


@_codec_provider
def _text_codec_provider(python_type):
    """
    Provides a codec that encodes/decodes a value to/from a SQLite TEXT. It unconditionally
    returns the codec, regardless of Python type. It should be the last provider in the
    providers list to serve as a catch-all.
    """

    str_codec = fondat.codec.get_codec(String, python_type)

    class TextCodec(SQLiteCodec):

        sql_type = "TEXT"

        @validate_arguments
        def encode(self, value: python_type) -> str:
            return str_codec.encode(value)

        @validate_arguments
        def decode(self, value: str) -> python_type:
            return str_codec.decode(value)

    return TextCodec()


class _Results(AsyncIterator[Any]):

    __slots__ = {"statement", "result", "rows", "codecs"}

    def __init__(self, statement, result, rows):
        self.statement = statement
        self.result = result
        self.rows = rows
        self.codecs = {
            k: get_codec(t)
            for k, t in typing.get_type_hints(result, include_extras=True).items()
        }

    def __aiter__(self):
        return self

    async def __anext__(self):
        row = await self.rows.__anext__()
        return self.result(**{k: self.codecs[k].decode(row[k]) for k in self.codecs})


@asynccontextmanager
async def _async_null_context():
    yield


class Database(fondat.sql.Database):
    """
    Manages access to a SQLite database.

    Parameter:
    • path: path to SQLite database file
    """

    __slots__ = {"path", "_conn", "_txn"}

    def __init__(self, path: str):
        super().__init__()
        self.path = path
        self._conn = contextvars.ContextVar("fondat_sqlite_conn", default=None)
        self._txn = contextvars.ContextVar("fondat_sqlite_txn", default=None)
        self._task = contextvars.ContextVar("fondat_sqlite_task", default=None)

    @asynccontextmanager
    async def connection(self):
        task = asyncio.current_task()
        if self._conn.get() and self._task.get() is task:
            yield  # connection already established
            return
        _logger.debug("open connection")
        self._task.set(task)
        connection = await aiosqlite.connect(self.path)
        connection.row_factory = sqlite3.Row
        self._conn.set(connection)
        try:
            yield
        finally:
            _logger.debug("close connection")
            self._conn.set(None)
            try:
                await connection.close()
            except Exception as e:
                _logger.exception("error closing connection")

    @asynccontextmanager
    async def transaction(self):
        txid = f"_{uuid.uuid4().hex}"
        _logger.debug("transaction begin %s", txid)
        token = self._txn.set(txid)

        async def commit():
            _logger.debug("transaction commit %s", txid)
            await connection.execute(f"RELEASE SAVEPOINT {txid};")

        async def rollback():
            _logger.debug("transaction rollback %s", txid)
            await connection.execute(f"ROLLBACK TO SAVEPOINT {txid};")

        async with self.connection():
            connection = self._conn.get()
            await connection.execute(f"SAVEPOINT {txid};")
            try:
                yield
            except GeneratorExit:  # explicit cleanup of asynchronous generator
                await commit()
            except Exception:
                await rollback()
                raise
            else:
                await commit()
            finally:
                self._txn.reset(token)

    async def execute(
        self,
        statement: Expression,
        result: type = None,
    ) -> AsyncIterator[Any] | None:
        if not self._txn.get():
            raise RuntimeError("transaction context required to execute statement")
        text = []
        args = []
        for fragment in statement:
            match fragment:
                case str():
                    text.append(fragment)
                case Param():
                    text.append("?")
                    args.append(get_codec(fragment.type).encode(fragment.value))
                case _:
                    raise ValueError(f"unexpected fragment: {fragment}")
        results = await self._conn.get().execute("".join(text), args)
        if result is not None:  # expecting a result
            return _Results(statement, result, results.__aiter__())

    def sql_type(self, type: Any) -> str:
        return get_codec(type).sql_type



"""Module to store resource items in files."""

import logging
import mimetypes

from collections.abc import Iterable
from fondat.codec import Binary, String, get_codec
from fondat.error import InternalServerError, NotFoundError
from fondat.http import AsBody
from fondat.pagination import Page
from fondat.resource import operation, resource
from fondat.security import Policy
from fondat.stream import Stream
from pathlib import Path
from typing import Annotated, Any
from urllib.parse import quote, unquote


_logger = logging.getLogger(__name__)


_content_encoding_map = {
    "br": "application/x-br",
    "bzip2": "application/x-bzip2",
    "compress": "application/x-compress",
    "gzip": "application/gzip",
    "xz": "application/x-xz",
}


def _content_type(url: str) -> str:
    content_type, content_encoding = mimetypes.guess_type(url)
    if content_encoding:
        return _content_encoding_map.get(content_encoding, content_encoding)
    if content_type:
        return content_type
    return "application/octet-stream"


class _ReadFileStream(Stream):
    def __init__(self, path: Path, block_size: int = 131072):
        super().__init__(
            content_type=_content_type(path.name), content_length=path.stat().st_size
        )
        self.path = path
        self.block_size = block_size
        self.position = 0

    async def __anext__(self) -> bytes:
        if not self.path:
            raise StopAsyncIteration
        with self.path.open("rb") as file:
            file.seek(self.position)
            block = file.read1(self.block_size)
        if len(block) == 0:
            raise StopAsyncIteration
        self.position += len(block)
        return block


def _tmp_path(path: Path):
    return path.with_name(f"{path.name}.__tmp__")


def _stream_resource_class(
    writeable: bool,
    publish: bool,
    policies: Iterable[Policy],
):
    @resource
    class StreamResource:
        def __init__(self, path: Path):
            self.path = path

        @operation(publish=publish, policies=policies)
        async def get(self) -> Stream:
            """Read resource."""
            if not self.path.is_file():
                raise NotFoundError
            return _ReadFileStream(self.path)

        if writeable:

            @operation(publish=publish, policies=policies)
            async def put(self, value: Annotated[Stream, AsBody]):
                """Write resource."""
                tmp = _tmp_path(self.path)
                try:
                    with tmp.open("xb") as file:
                        async for block in value:
                            file.write(block)
                    tmp.replace(self.path)
                except Exception as e:
                    raise InternalServerError from e

            @operation(publish=publish, policies=policies)
            async def delete(self):
                """Delete resource."""
                try:
                    self.path.unlink()
                except FileNotFoundError:
                    raise NotFoundError
                except Exception as e:
                    raise InternalServerError from e

    StreamResource.__qualname__ = "StreamResource"
    return StreamResource


def _file_resource_class(
    value_type: type,
    compress: Any,
    writeable: bool,
    publish: bool,
    policies: Iterable[Policy],
):

    if value_type is Stream:
        if compress is not None:
            raise TypeError("file resources does not support compression of streams")
        return _stream_resource_class(writeable, publish, policies)

    codec = get_codec(Binary, value_type)

    @resource
    class FileResource:
        def __init__(self, path: Path):
            self.path = path

        @operation(publish=publish, policies=policies)
        async def get(self) -> value_type:
            """Read resource."""
            try:
                with self.path.open("rb") as file:
                    content = file.read()
                if compress:
                    content = compress.decompress(content)
                return codec.decode(content)
            except FileNotFoundError:
                raise NotFoundError
            except Exception as e:
                raise InternalServerError from e

        if writeable:

            @operation(publish=publish, policies=policies)
            async def put(self, value: Annotated[value_type, AsBody]):
                """Write resource."""
                content = codec.encode(value)
                if compress:
                    content = compress.compress(content)
                tmp = _tmp_path(self.path)
                try:
                    with tmp.open("xb") as file:
                        file.write(content)
                    tmp.replace(self.path)
                except Exception as e:
                    raise InternalServerError from e

            @operation(publish=publish, policies=policies)
            async def delete(self):
                """Delete resource."""
                try:
                    self.path.unlink()
                except FileNotFoundError:
                    raise NotFoundError

    FileResource.__qualname__ = "FileResource"
    return FileResource


def file_resource(
    path: Path | str,
    value_type: type = Stream,
    compress: Any = None,
    writeable: bool = False,
    publish: bool = True,
    policies: Iterable[Policy] | None = None,
) -> type:
    """
    Return a new resource that manages a file.

    Parameters:
    • path: path of file
    • value_type: type of value stored in file
    • compress: algorithm to compress and decompress file content
    • writeable: can file be written or deleted
    • publish: publish the operation in documentation
    • policies: security policies to apply to file operations

    Compression algorithm is any object or module that exposes callable "compress" and
    "decompress" attributes. Examples: bz2, gzip, lzma, zlib.
    """
    if isinstance(path, str):
        path = Path(path)
    return _file_resource_class(value_type, compress, writeable, publish, policies)(path)


def _limit(requested):
    upper = 1000
    if not requested or requested < 0:
        return upper
    return min(requested, upper)


def directory_resource(
    path: Path | str,
    key_type: type = str,
    value_type: type = Stream,
    extension: str | None = None,
    compress: Any = None,
    writeable: bool = False,
    index: bool = True,
    publish: bool = True,
    policies: Iterable[Policy] | None = None,
) -> type:
    """
    Return a new resource that manages files in a directory.

    Parameters:
    • path: path to directory where files are stored
    • key_type: type of key to identify file
    • value_type: type of value stored in each file
    • extenson: filename extension to append (including dot)
    • compress: algorithm to compress and decompress file content
    • writeable: can files be written or deleted
    • index: provide index of files with get method
    • publish: publish the operation in documentation
    • policies: Security requirements to apply to all operations

    Compression algorithm is any object or module that exposes callable "compress" and
    "decompress" attributes. Examples: bz2, gzip, lzma, zlib. Compression is not supported for
    value_type of Stream.

    The index parameter can be one of the following:
    • key_type: identifies a file in the directory to provide as the index
    • None: generates a paginated list of file keys as the index
    """

    _path = (Path(path) if isinstance(path, str) else path).expanduser()

    if not _path.is_dir():
        raise FileNotFoundError(f"directory not found: {_path}")

    codec = get_codec(String, key_type)

    FileResource = _file_resource_class(value_type, compress, writeable, publish, policies)

    @resource
    class DirectoryResource:
        def __getitem__(self, key: key_type) -> FileResource:
            return FileResource(
                _path.joinpath(
                    f"{quote(codec.encode(key), safe='')}{extension if extension else ''}"
                )
            )

        if index:

            @operation(publish=publish, policies=policies)
            async def get(
                self,
                limit: int | None = None,
                cursor: bytes | None = None,
            ) -> Page[key_type]:
                """Return paginated list of file keys."""
                limit = _limit(limit)
                if cursor is not None:
                    cursor = cursor.decode()
                try:
                    if not extension:
                        names = sorted(
                            entry.name for entry in _path.iterdir() if entry.is_file()
                        )
                    else:
                        names = sorted(
                            entry.name[: -len(extension)]
                            for entry in _path.iterdir()
                            if entry.is_file() and entry.name.endswith(extension)
                        )
                except FileNotFoundError as fnfe:
                    raise InternalServerError from fnfe
                page = Page[key_type](items=[], cursor=None)
                for (counter, name) in enumerate(names, 1):
                    if cursor is not None:
                        if name <= cursor:
                            continue
                        cursor = None
                    try:
                        page.items.append(codec.decode(unquote(name)))
                    except ValueError:
                        continue  # ignore name that cannot be decoded
                    if len(page.items) == limit and counter < len(names):
                        page.cursor = name.encode()
                        break
                return page

    DirectoryResource.__qualname__ = "DirectoryResource"

    return DirectoryResource()



"""Module to store resource items in memory."""

from __future__ import annotations

import threading

from collections import namedtuple
from collections.abc import Iterable
from copy import deepcopy
from fondat.error import InternalServerError, NotFoundError
from fondat.http import AsBody
from fondat.resource import mutation, operation, resource
from fondat.security import Policy
from fondat.types import affix_type_hints
from time import time
from typing import Annotated


_Item = namedtuple("Item", "value,time")

_Oldest = namedtuple("Oldest", "key,time")


def memory_resource(
    key_type: type,
    value_type: type,
    size: int | None = None,
    evict: bool = False,
    expire: int | float | None = None,
    publish: bool = True,
    policies: Iterable[Policy] | None = None,
):
    """
    Return a new resource that stores items in memory.

    Parameters:
    • key_type: type for the key for each item
    • value_type: type for value stored in each item
    • size: maximum number of items to store  [unlimited]
    • evict: should oldest item be evicted to make room for a new item
    • expire: expire time for each value in seconds  [unlimited]
    • publish: publish the operation in documentation
    • policies: security policies to apply to all operations
    """

    @resource
    class MemoryResource:
        def __init__(self):
            self.storage = {}
            self.lock = threading.Lock()

        def __getitem__(self, key: key_type) -> Item:
            return Item(self, key)

        @operation(publish=publish, policies=policies)
        async def get(self) -> list[key_type]:
            """Return list of item keys."""
            now = time()
            with self.lock:
                return [
                    key
                    for key, item in self.storage.items()
                    if not self.expire or item.time + self.expire <= now
                ]

        @mutation(publish=publish, policies=policies)
        async def clear(self) -> None:
            """Remove all items."""
            with self.lock:
                self.storage.clear()

    MemoryResource.key_type = key_type
    MemoryResource.value_type = value_type
    MemoryResource.size = size
    MemoryResource.evict = evict
    MemoryResource.expire = expire

    @resource
    class Item:
        def __init__(self, container: MemoryResource, key: key_type):
            self.container = container
            self.key = key

        @operation(publish=publish, policies=policies)
        async def get(self) -> value_type:
            """Read item."""
            item = self.container.storage.get(self.key)
            if item is None or (
                self.container.expire and time() > item.time + self.container.expire
            ):
                raise NotFoundError
            return deepcopy(item.value)

        @operation(publish=publish, policies=policies)
        async def put(self, value: Annotated[value_type, AsBody]) -> None:
            """Write item."""
            with self.container.lock:
                now = time()
                if self.container.expire:  # purge expired entries; pay the price on puts
                    for key in {
                        k
                        for k, v in self.container.storage.items()
                        if v.time + self.container.expire <= now
                    }:
                        del self.container.storage[key]
                while (
                    self.container.evict
                    and self.container.size
                    and len(self.container.storage) >= self.container.size
                ):  # evict oldest entry
                    oldest = None
                    for _key, _item in self.container.storage.items():
                        if not oldest or _item.time < oldest.time:
                            oldest = _Oldest(_key, _item.time)
                    if oldest:
                        del self.container.storage[oldest.key]
                if self.container.size and len(self.container.storage) >= self.container.size:
                    raise InternalServerError from RuntimeError("item size limit reached")
                self.container.storage[self.key] = _Item(value, now)

        @operation(publish=publish, policies=policies)
        async def delete(self):
            """Delete item."""
            await self.get()  # ensure item exists and has not expired
            with self.container.lock:
                self.container.storage.pop(self.key, None)

    affix_type_hints(MemoryResource, localns=locals())
    affix_type_hints(Item, localns=locals())

    MemoryResource.__qualname__ = "MemoryResource"

    return MemoryResource()
